Package: scDesign3
Type: Package
Title: A unified framework of realistic in silico data generation and
        statistical model inference for single-cell and spatial omics
Version: 1.10.0
Authors@R: 
    c(person("Dongyuan", "Song", , "dongyuansong@ucla.edu", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-1114-1215")),
    person("Qingyang", "Wang", , "qw802@g.ucla.edu",  role = c("aut"),
           comment = c(ORCID = "0000-0002-1051-609X")),
    person("Chenxin", "Jiang", , "cflorajiang@g.ucla.edu",  role = c("aut"),
           comment = c(ORCID = "0009-0005-7369-4116")))
Description: We present a statistical simulator, scDesign3, to generate realistic single-cell and spatial omics data, including various cell states, experimental designs, and feature modalities, by learning interpretable parameters from real data. Using a unified probabilistic model for single-cell and spatial omics data, scDesign3 infers biologically meaningful parameters; assesses the goodness-of-fit of inferred cell clusters, trajectories, and spatial locations; and generates in silico negative and positive controls for benchmarking computational tools.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: false
Depends: R (>= 4.3.0)
Imports: dplyr, tibble, stats, methods, mgcv, gamlss, gamlss.dist,
        SummarizedExperiment, SingleCellExperiment, mclust, mvtnorm,
        parallel, pbmcapply, umap, ggplot2, irlba, viridis,
        BiocParallel, matrixStats, Matrix, sparseMVN, coop
Suggests: mvnfast, igraph, rvinecopulib, knitr, rmarkdown, testthat (>=
        3.0.0), RefManageR, sessioninfo, BiocStyle
biocViews: Software, SingleCell, Sequencing, GeneExpression, Spatial
URL: https://github.com/SONGDONGYUAN1994/scDesign3
BugReports: https://github.com/SONGDONGYUAN1994/scDesign3/issues
RoxygenNote: 7.3.3
Config/testthat/edition: 3
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/scDesign3
git_branch: RELEASE_3_23
git_last_commit: c1f38d2
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-29 00:27:40 UTC; biocbuild
Author: Dongyuan Song [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-1114-1215>),
  Qingyang Wang [aut] (ORCID: <https://orcid.org/0000-0002-1051-609X>),
  Chenxin Jiang [aut] (ORCID: <https://orcid.org/0009-0005-7369-4116>)
Maintainer: Dongyuan Song <dongyuansong@ucla.edu>
