# Generated by roxygen2: do not edit by hand

S3method(predict,gamlss)
export(ba)
export(construct_data)
export(extract_para)
export(fit_copula)
export(fit_marginal)
export(ga)
export(gamlss.ba)
export(gamlss.ga)
export(perform_lrt)
export(plot_reduceddim)
export(scdesign3)
export(simu_new)
export(sparse_cov)
import(gamlss)
import(ggplot2)
import(mclust)
